// ignore_for_file: avoid_print

import 'dart:async';

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:test/test.dart';

class UserTest {}

Future<void> luongTests() async {
  test('20210929-test1', () async {
    await StorageMapper.initDatabaseExpander();
    StorageController geigerToolboxStorageController =
        GenericController('MI-Cyberrange', SqliteMapper('dbFileName.sqlite'));
    await geigerToolboxStorageController.close();
    // File('dbFileName.sqlite').deleteSync();
  });
  test('20210929-test2', () async {
    // init
    await StorageMapper.initDatabaseExpander();
    StorageController geigerToolboxStorageController =
        GenericController('MI-Cyberrange', DummyMapper('testowner'));

    // sample code
    var newScore = 123;
    var currentUserLevel = 123;
    try {
      Node node = await geigerToolboxStorageController.get(':score-node');
      // this is bad coding as it depends on a full transaction database state. Introduced addOrUpdateValue as drop in replacement.
      node.addValue(NodeValueImpl('score', '$newScore'));
      node.addValue(NodeValueImpl('level', '$currentUserLevel'));
      await geigerToolboxStorageController.update(node);
    } catch (e) {
      print('Exception while creating node');
      Node node = NodeImpl('score-node', 'testowner');
      await geigerToolboxStorageController.add(node);
      node.addValue(NodeValueImpl('score', '$newScore'));
      node.addValue(NodeValueImpl('level', '$currentUserLevel'));
      await geigerToolboxStorageController.update(node);
    }
  });
}

class Event {
  final EventType _event;
  final Node? _old;
  final Node? _new;

  Event(this._event, this._old, this._new);

  EventType get type => _event;

  Node? get oldNode => _old;

  Node? get newNode => _new;

  @override
  String toString() {
    return '${type.toString()} ${oldNode.toString()}=>${newNode.toString()}';
  }
}

class CustomStorageListener implements StorageListener {
  List<Event> events = <Event>[];

  @override
  void gotStorageChange(EventType event, Node? oldNode, Node? newNode) {
    Event e = Event(event, oldNode, newNode);
    events.add(e);
    print('got event ${e.toString()}');
  }

  Future<List<Event>> getNumEvents(int num, [int timeout = 1000000]) async {
    int start = DateTime.now().millisecondsSinceEpoch;
    List<Event> ret = await Future.doWhile(() =>
            events.length < num &&
            start + 1000 * timeout > DateTime.now().millisecondsSinceEpoch)
        .then((value) => events);
    if (events.length < num) {
      throw TimeoutException('Timeout reached while waiting for $num events');
    }
    return ret;
  }
}

Future<void> nosaTests() async {
  test(
      '20211012-flutter Github Issue 16 (https://github.com/cyber-geiger/toolbox-storage/issues/16)',
      () async {
    StorageController _storageController =
        GenericController("Example", SqliteMapper("database.db"));
    await _storageController.zap();
    CustomStorageListener l = CustomStorageListener();
    SearchCriteria s = SearchCriteria(searchPath: ':');
    _storageController.registerChangeListener(l, s);
    Node n = NodeImpl(':hello', 'testowner');
    expect(await s.evaluate(n), isTrue,
        reason: 'node fails to evaluate successfully');
    List<Event> e = await _storageController
        .add(n)
        .then((_) async => await l.getNumEvents(1, 2));
    expect(e.length, 1, reason: 'Did not get expected number of results');
  });
  test('20211119-flutter Github issue #18', () async {
    //initialize database
    StorageController _storageController =
        GenericController("Example", SqliteMapper("example.db"));
    await _storageController.zap();

    Future<void> setGlobalThreatsNode() async {
      List<String> threats = ["phishing", "malware"];
      List<String> threatid = ["t1", "t2"];
      try {
        for (int i = 0; i < threats.length; i++) {
          Node _node =
              await _storageController.get(':Global:threats:${threatid[i]}');
          //create a NodeValue
          NodeValue threatNodeValueName = NodeValueImpl("name", threats[i]);
          await _node.addOrUpdateValue(threatNodeValueName);
          await _storageController.update(_node);
        }
      } on StorageException {
        //print(":Global:threats not found");
        try {
          Node threatsNode = NodeImpl(":Global:threats", "testowner");
          await _storageController.addOrUpdate(threatsNode);

          for (int i = 0; i < threats.length; i++) {
            Node threatIdNode =
                NodeImpl(":Global:threats:${threatid[i]}", "testowner");
            //create :Global:threats:$threatId
            await _storageController.addOrUpdate(threatIdNode);
            //create a NodeValue
            NodeValue threatNodeValueName = NodeValueImpl("name", threats[i]);
            await threatIdNode.addOrUpdateValue(threatNodeValueName);
            await _storageController.update(threatIdNode);
          }
        } catch (e, st) {
          print(
              "*********************************************************************");
          print(e);
          print(st);
          fail('failed due to unexpected exception');
        }
      }
    }

    Future<List<String>> getThreats({String language = "en"}) async {
      List<String> t = [];
      try {
        Node _node = await _storageController.get(":Global:threats");

        //return _node!.getChildNodesCsv();
        for (String threatId in await _node
            .getChildNodesCsv()
            .then((value) => value.split(','))) {
          Node threatNode =
              await _storageController.get(":Global:threats:$threatId");

          t.add(await threatNode
              .getValue("name")
              .then((value) => value!.getValue(language)!));
        }
      } on StorageException catch (e) {
        print(
            "*********************************************************************");
        print('$e');
        fail('Got Storage Exception while fetching list of threats');
      } on Error catch (e) {
        print(
            "*********************************************************************");
        print('$e');
        fail('Got Exception while fetching list of threats');
      }
      return t;
    }

    Future<List<String>> getListThreat() async {
      await setGlobalThreatsNode();
      return await getThreats();
    }

    print("Global Threats: ${await getListThreat()}");
  });
  test('20211119-flutter Github issue #18 (rewrite)', () async {
    //initialize database
    await StorageMapper.initDatabaseExpander();
    StorageController _storageController =
        GenericController("Example", SqliteMapper("example.db"));
    // Make sure we start off with a fresh DB
    await _storageController.zap();

    Future<void> setGlobalThreatsNode() async {
      Map<String, String> threats = {"t1": "phishing", "t2": "malware"};
      try {
        // Make sure that parent node exists
        Node threatsNode = NodeImpl(":Global:threats", "testowner");
        await _storageController.addOrUpdate(threatsNode);

        // create threats
        for (MapEntry<String, String> e in threats.entries) {
          // create node
          Node _node = NodeImpl(':Global:threats:${e.key}', 'testowner');
          await _node.addOrUpdateValue(NodeValueImpl("name", e.value));
          await _storageController.addOrUpdate(_node);
        }
      } on StorageException catch (e, st) {
        print(
            "*********************************************************************");
        print(e);
        print(st);
        fail('failed due to unexpected exception');
      }
    }

    Future<List<String>> getThreats({String language = "en"}) async {
      List<String> t = [];
      try {
        List<Node> _node = await _storageController
            .search(SearchCriteria(searchPath: ":Global:threats"));
        for (Node element in _node) {
          if (element.parentPath == ":Global:threats") {
            t.add((await element.getValue('name'))!.value);
          }
        }
      } on Error catch (e) {
        print(
            "*********************************************************************");
        print('$e');
        fail('Got Exception while fetching list of threats');
      }
      return t;
    }

    Future<List<String>> getListThreat() async {
      await setGlobalThreatsNode();
      return await getThreats();
    }

    print("Global Threats: ${await getListThreat()}");
  });
  test(
      '20211129 - Unhandled Exception: Null check operator used on a null value Error #24',
      () async {
    Future<StorageController?> initGeigerStorage() async {
      try {
        await StorageMapper.initDatabaseExpander();
        StorageController geigerToolboxStorageController =
            GenericController('owner', SqliteMapper('database.db'));
        return geigerToolboxStorageController;
      } catch (e) {
        print("Failed to connect to database");
        fail('unable to open database');
      }
    }

    Future<void> writeToGeigerStorage(
        StorageController geigerToolboxStorageController, String data) async {
      print('Trying to get the data node');
      String dataNodeName = ':dataNode';
      try {
        print('Found the data node - Going to write the data');
        Node node = await geigerToolboxStorageController.get(dataNodeName);
        await node.updateValue(NodeValueImpl('data', data));
        await geigerToolboxStorageController.update(node);
      } catch (e) {
        print(e.toString());
        print('Cannot find the data node - Going to create a new one');
        try {
          Node node = NodeImpl(dataNodeName, 'owner');
          node.visibility = Visibility.green;

          await geigerToolboxStorageController.addOrUpdate(node);
          await node.addOrUpdateValue(NodeValueImpl('data', data));
          await geigerToolboxStorageController.update(node);
        } catch (e2, s2) {
          print(e2.toString());
          print(s2.toString());
          fail('out of luck');
        }
      }
    }

    Future<String?> readDataFromGeigerStorage(
        StorageController geigerToolboxStorageController) async {
      String nodeDataName = ':dataNode';
      print('Trying to get the data node');
      try {
        print('Found the data node - Going to get the data');
        Node node = await geigerToolboxStorageController.get(nodeDataName);
        List<Node> nodes = await geigerToolboxStorageController
            .search(SearchCriteria(searchPath: nodeDataName));
        print("list of path node: $nodes");
        NodeValue? nValue = await node.getValue('data');
        if (nValue != null) {
          return nValue.value;
        } else {
          fail('no data found in node value');
        }
      } catch (e, s) {
        print(e.toString());
        print(s.toString());
        fail('Failed to retrieve the data node');
      }
    }

    try {
      StorageController? c = await initGeigerStorage();
      await writeToGeigerStorage(c!, "nosa");
      String? result = await readDataFromGeigerStorage(c);
      expect(result, 'nosa');
    } catch (e) {
      fail('failed due to unexpected exception');
    }
  });
  test(
      '20211129 - Unhandled Exception: Null check operator used on a null value Error #24 (rewrite)',
      () async {
    try {
      await StorageMapper.initDatabaseExpander();
      StorageController c =
          GenericController('owner', SqliteMapper('database.db'));
      ExtendedTimestamp.initializeTimestamp(c);
      c.addOrUpdateValue(':dataNode', NodeValueImpl('data', 'nosa'));
      String? result =
          (await c.getValue(':dataNode', 'data'))!.toSimpleString();
      expect(result, 'nosa');
    } catch (e) {
      fail('failed due to unexpected exception');
    }
  });
}

Future<void> main() async {
  await luongTests();
  await nosaTests();
}
